"""IOロードのインターフェース"""

from typing import List, Protocol, runtime_checkable
from jrdb_model import KaisaiData


@runtime_checkable
class IOLoader(Protocol):
    """IOロードのインターフェース"""

    def load(self) -> List[KaisaiData]:
        """開催データ一覧を取得する

        Returns:
            List[KaisaiData]: 開催データ一覧
        """



"""KaisaiDataを取得し、ndarrayに変換する"""

from typing import Callable, List, Optional
from jrdb_model import KaisaiData
from jra_tools.machine_learning.usecase.icreator import ICreator
from jra_tools.machine_learning.usecase.iquery import IQuery

CreatorFactory = Callable[[List[KaisaiData]], ICreator]


class Converter:
    """KaisaiDataをndarrayに変換するクラス"""

    def __init__(self, query: IQuery, factories: List[CreatorFactory]):
        assert isinstance(query, IQuery)
        self.query = query
        self.factories = factories

    def save(self, name: Optional[str] = None):
        """指定したファイル名でndarrayを保存する

        Args:
            name (str): ファイル名
        """
        kaisais = self.query.load()
        for factory in self.factories:
            creator = factory(kaisais)
            assert isinstance(creator, ICreator)
            creator.save(name if name is not None else self.query.period)



import numpy as np


class Labelref:
    def __init__(self):
        ur = np.zeros([18, 18])
        s0 = 1
        for i in range(18):
            for j in range(i + 1, 18):
                if i != j:
                    ur[i, j] = s0
                    s0 += 1
        self.umaren_ref = ur

        ut = np.zeros([18, 18])
        s1 = 1
        for i in range(18):
            for j in range(18):
                if i != j:
                    ut[i, j] = s1
                    s1 += 1
        self.umatan_ref = ut

        sp = np.zeros([18, 18, 18])
        s2 = 1
        for i in range(18):
            for j in range(i + 1, 18):
                for k in range(j + 1, 18):
                    sp[i, j, k] = s2
                    s2 += 1
        self.sanrenpuku_ref = sp

        st = np.zeros([18, 18, 18])
        s3 = 1
        for i in range(18):
            for j in range(18):
                if i != j:
                    for k in range(18):
                        if j != k and k != i:
                            st[i, j, k] = s3
                            s3 += 1
        self.sanrentan_ref = st

    def getUmarenNumbers(self, n):
        one = 0
        two = 0
        if n == 1:
            one = 1
            two = 2
        else:
            for i in range(18):
                for j in range(i + 1, 18):
                    if self.umaren_ref[i, j] == n:
                        one = i + 1
                        two = j + 1
                        break
        return (one, two)

    def getUmarenCatnumber(self, n, m):
        return int(self.umaren_ref[n - 1, m - 1])

    def getUmatanNumbers(self, n):
        one = 0
        two = 0
        if n == 1:
            one = 1
            two = 2
        else:
            for i in range(18):
                for j in range(18):
                    if self.umatan_ref[i, j] == n:
                        one = i + 1
                        two = j + 1
                        break
        return (one, two)

    def getUmatanCatnumber(self, n, m):
        return int(self.umatan_ref[n - 1, m - 1])

    def getSanrenpukuNumbers(self, n):
        one = 0
        two = 0
        three = 0
        if n == 1:
            one = 1
            two = 2
            three = 3
        else:
            for i in range(18):
                for j in range(i + 1, 18):
                    for k in range(j + 1, 18):
                        if self.sanrenpuku_ref[i, j, k] == n:
                            one = i + 1
                            two = j + 1
                            three = k + 1
                            break
        return (one, two, three)

    def getSanrenpukuCatnumber(self, n, m, l):
        return int(self.sanrenpuku_ref[n - 1, m - 1, l - 1])

    def getSanrentanNumbers(self, n):
        one = 0
        two = 0
        three = 0
        if n == 1:
            one = 1
            two = 2
            three = 3
        else:
            for i in range(18):
                for j in range(18):
                    if i != j:
                        for k in range(18):
                            if self.sanrentan_ref[i, j, k] == n:
                                one = i + 1
                                two = j + 1
                                three = k + 1
                                break
        return (one, two, three)

    def getSanrentanCatnumber(self, n, m, l):
        return int(self.sanrentan_ref[n - 1, m - 1, l - 1])



"""DB周りのユーティリティ"""

from typing import List
from functools import reduce
import numpy as np
from jra_tools import KaisaiData


def create_payout(kaisais: List[KaisaiData]) -> np.ndarray:
    """払戻データをndarrayに変換する

    Args:
        kaisais (list[KaisaiData]): 開催データ一覧

    Returns:
        np.ndarray: (レース数, 18)
    """
    payout_data = np.zeros([reduce(lambda x, y: x + len(y.races), kaisais, 0), 18])

    num_race = 0
    for kaisai in kaisais:
        for race in kaisai.races:
            payout_data[num_race, race.returninfo.win1_num - 1] = (
                race.returninfo.win1_ret
            )
            num_race += 1
    return payout_data


def save_payout(payout: np.ndarray, period: str):
    with open(f"payout_{period}.dump", "wb") as f:
        payout.dump(f)


def save_payout_from(kaisais: List[KaisaiData], period: str):
    """開催一覧から払戻金データをローカルに保存する

    Args:
        kaisais (List[KaisaiData]): 開催一覧
        period (str): ファイル名
    """
    save_payout(create_payout(kaisais), period)


class PayoutCreator:
    """kaisaisから払戻金を作るクラス"""

    def __init__(self, kaisais):
        self.kaisais = kaisais
        self.payout = create_payout(kaisais)

    def save(self, period: str):
        """payoutのndarrayをローカルに保存する

        Args:
            period (str): 拡張子を除いたファイル名
        """
        save_payout(self.payout, period)



"""トレーニングデータ取得用クエリ"""

from typing import List, Callable
from jrdb_model import KaisaiData
from jra_tools.machine_learning.interface_adapter.ispecified_loader import (
    SpecifiedLoader,
)


def _validate_year(year: int):
    assert 2000 <= year <= 2100, "2000年以降を指定してください"
    assert isinstance(year, int)


class TrainingKaisaiQuery:
    """トレーニングデータ用クエリ"""

    def __init__(
        self,
        start: int,
        end: int,
        loader_factory: Callable[[int, int], SpecifiedLoader],
    ):
        _validate_year(start)
        _validate_year(end)
        assert start <= end, "終了年は開始年以降にしてください"
        self.start = start
        self.end = end
        self.loader = loader_factory(self._start, self._end)

    def load(self) -> List[KaisaiData]:
        return self.loader.load()

    @property
    def _start(self):
        return int(f"{self.start}0101")

    @property
    def _end(self):
        return int(f"{self.end}1231")

    @property
    def period(self):
        """データセットの期間"""
        return (
            "data"
            if self.start == 2012 and self.end == 2018
            else f"{self.start}_{self.end}"
        )



"""推論データを生成する."""

from typing import List

from jrdb_model import (
    KaisaiData,
    PredictData,
)
from keras import Model

from jra_tools.machine_learning.interface_adapter.input_creator import InputCreator


class PredictFactory:
    """推論データ生成クラス."""

    def __init__(self, kaisais: List[KaisaiData], model: Model):
        for kaisai in kaisais:
            if not isinstance(kaisai, KaisaiData):
                raise TypeError("開催データではありません")
        self.kaisais = kaisais

        if not isinstance(model, Model):
            raise TypeError("Keras V3のモデルではありません")
        self.model = model

    def create(self) -> List[PredictData]:
        """推論データを生成する.

        Returns:
            List[PredictData]: 未登録の推論データ一覧

        """
        input_data = InputCreator(self.kaisais)
        preds = self.model.predict(input_data.x_data)
        w = 0
        entities = []
        for k in self.kaisais:
            for r in k.races:
                for h in r.racehorses:
                    hn = h.num - 1
                    pp1 = preds["activation_4"][w][hn]
                    pp2 = preds["activation_5"][w][hn]
                    pp3 = preds["activation_6"][w][hn]
                    rentai_rate = 1 - ((1 - pp1) * (1 - pp2))
                    fukusho_rate = 1 - ((1 - pp1) * (1 - pp2) * (1 - pp3))
                    entities.append(
                        PredictData(
                            racehorsekey=h.racehorsekey,
                            pp_icchaku=float(pp1),
                            pp_nichaku=float(pp2),
                            pp_sanchaku=float(pp3),
                            rentai_rate=rentai_rate,
                            fukusho_rate=fukusho_rate,
                            tansho_odds=1 / pp1,
                            fukusho_odds=1 / fukusho_rate,
                        )
                    )
                w += 1
        return entities



"""ツールのユースケース"""

from functools import partial
from jra_tools.machine_learning.infrastructure.mysql_loader import MysqlLoader
from jra_tools.machine_learning.interface_adapter.input_creator import InputCreator
from jra_tools.machine_learning.interface_adapter.payout_creator import PayoutCreator
from jra_tools.machine_learning.interface_adapter.training_kaisai_query import (
    TrainingKaisaiQuery,
)
from jra_tools.machine_learning.interface_adapter.quarter_kaisai_query import (
    QuarterKaisaiQuery,
)
from jra_tools.machine_learning.usecase.converter import Converter


def _create_training(start: int, end: int, only_input: bool):
    """指定した期間のトレーニングデータを作成する

    Args:
        start (int): YYYY
        end (int): YYYY

    Raises:
        RuntimeError: データのロードまたは保存中にエラーが発生
    """

    converter = Converter(
        TrainingKaisaiQuery(start, end, MysqlLoader),
        [InputCreator] if only_input else [InputCreator, PayoutCreator],
    )
    converter.save()


create_training_dataset = partial(_create_training, only_input=False)
create_training_input = partial(_create_training, only_input=True)


def _create_quarter(year: int, quarter: int, only_input: bool) -> None:
    """指定した四半期のデータセットをファイルに保存する

    Args:
        year (int): YYYY
        quarter (int): 1 - 4
        only_input (bool, optional): 入力データのみを作成するか. Defaults to True.
    """
    converter = Converter(
        QuarterKaisaiQuery(year, quarter, MysqlLoader),
        [InputCreator] if only_input else [InputCreator, PayoutCreator],
    )
    converter.save()


create_quarter_dataset = partial(_create_quarter, only_input=False)
create_quarter_input = partial(_create_quarter, only_input=True)



"""Yahoo競馬からスケジュールを確認する"""
from typing import List
import re
from datetime import datetime
import asyncio
import aiohttp
from bs4 import BeautifulSoup

YEAR = int(datetime.now().strftime("%Y"))
MONTH = int(datetime.now().strftime("%m"))


async def annual_schedule(year: int = YEAR) -> List[int]:
    """指定して年間スケジュールを取得する

    Args:
        year (int, optional): 取得する年. Defaults to YEAR.

    Returns:
        List[int]: 開催される日付一覧
    """
    tasks = [open_days(month, year) for month in range(1, 13)]
    months = await asyncio.gather(*tasks)
    return [day for month in months for day in month]


async def open_days(month: int = MONTH, year: int = YEAR) -> List[int]:
    """指定した年、月の開催一覧を取得する

    Args:
        month (int, optional): 月. Defaults to MONTH.
        year (int, optional): 年. Defaults to YEAR.

    Returns:
        List[int]: 開催される日付一覧
    """
    base_url = "https://sports.yahoo.co.jp/keiba/schedule/monthly"
    payload = {"year": str(year), "month": str(month)}
    soup = BeautifulSoup(await _fetch_html(base_url, payload), "lxml")
    return [int(f"{year:04d}{month:02d}{day:02d}") for day in _parse_to_days(soup)]


async def _fetch_html(url, payload):
    async with aiohttp.ClientSession() as session:
        async with session.get(url, params=payload) as response:
            return await response.text()


def _parse_to_days(soup: BeautifulSoup) -> List[int]:
    return list(
        set(
            [
                int(re.sub("\\D", "", s.contents[0]))
                for s in soup.select(".hr-tableSchedule__data--date")
            ]
        )
    )



"""カテゴリ変数を取得する."""

from typing import List

import numpy as np
from jrdb_model import BangumiData, KaisaiData, RacehorseData

from jra_tools.machine_learning.entity.jrdbdummies import CategoryGetter


def get_category_data(kaisais: List[KaisaiData]) -> np.ndarray:
    """開催データからカテゴリデータを取得する.

    Args:
        kaisais (List[KaisaiData]): 開催データ

    Returns:
        np.ndarray: 開催データマトリクス

    """
    return _convert_to_matrix(_get_categories(kaisais))


def _convert_to_matrix(categories: List[np.ndarray]) -> np.ndarray:
    matrix = np.zeros([len(categories), 18, len(categories[0][0])])
    for race_num, race in enumerate(categories):
        for horse_num, horse in enumerate(race):
            matrix[race_num][horse_num] = horse
    return matrix


def _get_categories(kaisais: List[KaisaiData]) -> List[np.ndarray]:
    return [
        _get_race_categories(kaisai, race)
        for kaisai in kaisais
        for race in kaisai.races
    ]


def _get_race_categories(kaisai: KaisaiData, race: BangumiData) -> np.ndarray:
    for horse in sorted(race.racehorses, key=lambda h: h.racehorsekey):
        if horse.num == 1:
            dummies = _get_category(kaisai, race, horse)
        else:
            dummies = np.vstack((dummies, _get_category(kaisai, race, horse)))
    return dummies


def _get_category(
    kaisai: KaisaiData, race: BangumiData, horse: RacehorseData
) -> np.ndarray:
    cg = CategoryGetter()
    return np.hstack((cg.getWaku(horse.waku),))


def _filter_str_to_int(value):
    return 0 if isinstance(value, str) else value



"""クリエイターインターフェース"""

from typing import List, Protocol, runtime_checkable
from jrdb_model import KaisaiData


@runtime_checkable
class ICreator(Protocol):
    """kaisaisをndarrayに変換するインターフェース"""

    kaisais: List[KaisaiData]

    def save(self, name: str):
        """ローカルにデータセットを保存する

        Args:
            name (str): 拡張子を除いたファイル名
        """



import csv

from jrdb_model import UmarenOddsData, WakurenOddsData, WideOddsData, create_app, db


class BaoImporter:
    def __init__(self, kaisais):
        self.kaisaikeys = [
            {
                "course": self.getCourseCode(kaisai.kaisaikey),
                "ymd": kaisai.ymd,
                "key": kaisai.kaisaikey,
            }
            for kaisai in kaisais
        ]
        self.racekeys = [race.racekey for kaisai in kaisais for race in kaisai.races]

    def getKaisaiInfo(self, kaisaikey):
        return {
            "ba": kaisaikey[0:2],
            "year": kaisaikey[2:4],
            "kai": kaisaikey[4:5],
            "day": kaisaikey[5:6],
        }

    def getCourseCode(self, kaisaikey):
        return kaisaikey[0:2]

    def getRaceKey(self, bao_info):
        if self.kaisaikeys[0]["ymd"] <= bao_info["ymd"] <= self.kaisaikeys[-1]["ymd"]:
            for key in self.kaisaikeys:
                if (
                    key["ymd"] == bao_info["ymd"]
                    and key["course"] == bao_info["bacode"]
                ):
                    return key["key"] + bao_info["no"]

    def getBaoCodeInfo(self, race_code):
        year_header = "19" if race_code[0] == "1" else "20"
        return {
            "year": year_header + race_code[1:3],
            "month": race_code[3:5],
            "day": race_code[5:7],
            "ymd": int(year_header + race_code[1:7]),
            "bacode": race_code[7:9],
            "no": race_code[9:11],
        }

    def getUmarenOdds(self, odds):
        odds_dict = {}
        for i, odd_str in enumerate(odds):
            for j in range(0, len(odd_str), 6):
                odds_dict.update(
                    {
                        f"{i + 1}-{i + int((j + 6) / 6) + 1}": int(odd_str[j : j + 6])
                        / 10
                    }
                )
        return odds_dict

    def getWideOdds(self, odds):
        odds_dict = {}
        for i, odd_str in enumerate(odds):
            for j in range(0, len(odd_str), 10):
                odds_dict.update(
                    {
                        f"{i + 1}-{i + int((j + 10) / 10) + 1}": {
                            "min": int(odd_str[j : j + 5]) / 10,
                            "max": int(odd_str[j + 5 : j + 10]) / 10,
                        }
                    }
                )
        return odds_dict

    def getWakurenOdds(self, odds_str):
        odds_dict = {}
        pos = 0
        for i in range(8):
            for j in range(i, 8):
                odd = odds_str[pos : pos + 5].strip()
                odd = int(odd) / 10 if odd else None
                odds_dict.update({f"{i + 1}-{j + 1}": odd})
                pos += 5
        return odds_dict

    def convertRaceCodeToRaceKey(self, raceCode):
        bao_info = self.getBaoCodeInfo(raceCode)
        return self.getRaceKey(bao_info)

    def importUmarenWideCsv(self, fileName):
        with open(fileName) as f:
            reader = csv.reader(f)
            header = next(reader)
            app = create_app()
            with app.app_context():
                for row in reader:
                    racekey = self.convertRaceCodeToRaceKey(str(int(float(row[2]))))
                    if racekey:
                        odd = (
                            UmarenOddsData() if "umaren" in fileName else WideOddsData()
                        )
                        odd.racekey = racekey
                        odd.data_kbn = self._parseInt(row[0])
                        odd.registered_horses = self._parseInt(row[4])
                        odd.ran_horses = self._parseInt(row[5])
                        odd.sold_flg = self._parseInt(row[6])
                        odd.all_odds = (
                            self.getUmarenOdds(row[7].split())
                            if "umaren" in fileName
                            else self.getWideOdds(row[7].split())
                        )
                        odd.sum_of_all_bought_count = self._parseInt(row[8])
                        if racekey in self.racekeys:
                            db.session.add(odd)
                db.session.commit()

    def importWakurenCsv(self, fileName):
        with open(fileName) as f:
            reader = csv.reader(f)
            header = next(reader)
            with app.app_context():
                for row in reader:
                    racekey = self.convertRaceCodeToRaceKey(str(int(float(row[2]))))
                    if racekey:
                        odd = WakurenOddsData()
                        odd.racekey = racekey
                        odd.data_kbn = self._parseInt(row[0])
                        odd.registered_horses = self._parseInt(row[4])
                        odd.ran_horses = self._parseInt(row[5])
                        odd.sold_flg = self._parseInt(row[8])
                        odd.all_odds = self.getWakurenOdds(row[12])
                        odd.sum_of_all_bought_count = self._parseInt(row[15])
                        if racekey in self.racekeys:
                            db.session.add(odd)
                db.session.commit()

    def _parseInt(self, str):
        return 0 if str == "" else str



"""推論に必要なデータセットを作成する"""

from typing import List
import numpy as np
from jrdb_model import KaisaiData
from jra_tools.machine_learning.usecase.training_tool import (
    create_score_data_matrix,
    standardize,
)
from jra_tools.machine_learning.usecase.category_data import get_category_data


class InputCreator:
    """KaisaiDataを元にndarrayを作成する"""

    def __init__(self, kaisais: List[KaisaiData]):
        self.kaisais = kaisais
        self.x_data = create_input_data(kaisais)

    def save(self, name: str) -> None:
        """入力データを永続化する

        Args:
            name (str): ファイル名。拡張子も含めること。
        """
        save_input_data(self.x_data, name)


def save_input_data(x_data: np.ndarray, name: str) -> None:
    with open(f"x_{name}.dump", mode="wb") as f:
        x_data.dump(f)


def create_input_data(kaisais: List[KaisaiData]) -> np.ndarray:
    """開催一覧から推論モデル用入力データを作成する
    Args:
        kaisais (List[KaisaiData]): 開催一覧

    Returns:
        np.ndarray: 推論モデルの入力データ
    """
    score_data = standardize(create_score_data_matrix(kaisais))
    cd = get_category_data(kaisais)
    return np.concatenate((score_data, cd), axis=2)


def save_input_data_from(kaisais: List[KaisaiData], name: str):
    """開催一覧から入力データのndarrayをローカルに保存する

    Args:
        kaisais (List[KaisaiData]): 開催一覧
        name (str): ファイル名
    """
    save_input_data(create_input_data(kaisais), name)



"""クエリのインターフェース"""

from typing import List, Protocol, runtime_checkable
from jrdb_model import KaisaiData


@runtime_checkable
class IQuery(Protocol):
    """クエリのインターフェース"""

    @property
    def period(self) -> str:
        """取得する期間"""

    def load(self) -> List[KaisaiData]:
        """開催データ一覧を取得する

        Returns:
            List[KaisaiData]: 開催データ一覧
        """



"""開催データに紐づくデータを結合した状態でロードする."""

import datetime as dt
from dataclasses import dataclass
from typing import List, Tuple

from jrdb_model import KaisaiData, create_app
from sqlalchemy.orm import joinedload


def get_kaisais(start: int, end: int) -> List[KaisaiData]:
    """指定した期間の開催データを取得する.

    Args:
        start (int): YYYYMMDD
        end (int): YYYYMMDD

    Returns:
        list[KaisaiData]: 開催データ一覧

    """
    app = create_app()
    with app.app_context():
        return (
            KaisaiData.query.filter(KaisaiData.ymd >= start, KaisaiData.ymd <= end)
            .options(joinedload("*"))
            .all()
        )


@dataclass
class MysqlLoader:
    """MySQLからエンティティを取得する."""

    start: int
    end: int

    def load(self) -> List[KaisaiData]:
        return load(self.start, self.end)

    @property
    def period(self) -> str:
        return f"{self.start}_{self.end}"


def load(start: int, end: int) -> List[KaisaiData]:
    """データを1年毎にまとめてロードして、マージ結果を取得する.

    Args:
        start (int): YYYYMMDD
        end (int): YYYYMMDD

    Returns:
        List[KaisaiData]: 開催データ一覧

    """
    return [
        kaisai
        for start_date, end_date in _periods(_to_date(start), _to_date(end))
        for kaisai in get_kaisais(_to_int(start_date), _to_int(end_date))
    ]


def _periods(s: dt.datetime, e: dt.datetime) -> List[Tuple[dt.datetime, dt.datetime]]:
    if s.year < e.year:
        return [
            (s, dt.datetime(year=s.year, month=12, day=31)),
            *[_period(i) for i in range(s.year + 1, e.year)],
            (dt.datetime(year=e.year, month=1, day=1), e),
        ]
    if s.year == e.year and s <= e:
        return [(s, e)]
    raise RuntimeError("開始日 <= 終了日となるように入力してください")


def _period(year: int) -> Tuple[dt.datetime, dt.datetime]:
    return (
        dt.datetime(year=year, month=1, day=1),
        dt.datetime(year=year, month=12, day=31),
    )


def _to_date(date: int) -> dt.datetime:
    return dt.datetime.strptime(str(date), "%Y%m%d")


def _to_int(date: dt.datetime) -> int:
    return int(date.strftime("%Y%m%d"))



from jrdb_model import KaisaiData, create_app

from . import schedule


def missing_days(year):
    open_days = schedule.annual_schedule(year)
    return sorted(set(open_days) - set(db_days(year)))


def db_days(year):
    start_date = int(f"{year}0101")
    end_date = int(f"{year}1231")
    app = create_app()
    with app.app_context():
        kaisais = KaisaiData.query.filter(
            KaisaiData.ymd >= start_date, KaisaiData.ymd <= end_date
        ).all()
    return sorted(set([k.ymd for k in kaisais]))



"""四半期データ用クエリ"""

from typing import List, Callable
from jrdb_model import KaisaiData
from jra_tools.machine_learning.interface_adapter.ispecified_loader import (
    SpecifiedLoader,
)


class QuarterKaisaiQuery:
    """四半期データ用クエリ"""

    def __init__(
        self,
        year: int,
        quarter: int,
        loader_factory: Callable[[int, int], SpecifiedLoader],
    ):
        assert 2000 <= year <= 2100, "2000年以降を指定してください"
        assert isinstance(year, int)
        self.year = year
        assert 1 <= quarter <= 4, "1 - 4を指定してください"
        self.quarter = quarter
        self.loader = loader_factory(self._start, self._end)

    def load(self) -> List[KaisaiData]:
        return self.loader.load()

    @property
    def _start(self) -> int:
        return int(f"{self.year}{self._start_month:02}01")

    @property
    def _end(self) -> int:
        return int(
            f"{self.year}{self._end_month:02}{30 if self._end_month in [2,4,6,9,11] else 31}"
        )

    @property
    def period(self) -> str:
        """データセットの期間"""
        return f"{self.year}_{self._start_month:02}_{self._end_month:02}"

    @property
    def _start_month(self) -> int:
        return int(12 / 4 * (self.quarter - 1) + 1)

    @property
    def _end_month(self) -> int:
        return self._start_month + 2



"""パッケージから直接インポート可能なモジュール"""

from .database.bao_importer import *
from .database.db_checker import *
from .database.schedule import *
from .database.predict_factory import *
from .database.predict_race_factory import *

from .machine_learning.entity import *
from .machine_learning.usecase import *
from .machine_learning.interface_adapter import *
from .machine_learning.interface_adapter.input_creator import *
from .machine_learning.interface_adapter.payout_creator import *
from .machine_learning.interface_adapter.quarter_kaisai_query import *
from .machine_learning.interface_adapter.training_kaisai_query import *
from .machine_learning.infrastructure import *



"""推論レースデータを生成する"""
from typing import List
from dataclasses import dataclass
from jrdb_model import (
    PredictRaceData,
    KaisaiData,
    BangumiData,
    RacehorseData,
)


@dataclass
class PredictRaceFactory:
    """推論レースデータファクトリクラス"""

    kaisais: List[KaisaiData]

    def create(self) -> List[PredictRaceData]:
        """
        開催データ一覧から推論レースデータを生成する
        推論データが登録済みでないと失敗する

        Returns:
            List[PredictRaceData]: 推論レースデータ一覧
        """
        return [
            PredictRaceData(
                racekey=race.racekey,
                umaren=umarenOddses(race),
                wide=wideOddses(race),
                wakuren=wakurenOddses(race),
            )
            for kaisai in self.kaisais
            for race in kaisai.races
        ]


def umarenOddses(race: BangumiData):
    return oddses(race, "umaren")


def umarenOdds(horse1, horse2):
    p1 = 1 / umatanOdds(horse1, horse2)
    p2 = 1 / umatanOdds(horse2, horse1)
    p = p1 + p2
    return 1 / p


def umatanOdds(horse1, horse2):
    p1 = horse1.predict.pp_icchaku
    p2 = horse2.predict.pp_nichaku / (1 - horse1.predict.pp_nichaku)
    p = p1 * p2
    return 1 / p


def wideOddses(race: BangumiData):
    return oddses(race, "wide")


def oddses(race: BangumiData, baken: str):
    odds_dict = {}
    for i, horse1 in enumerate(race.racehorses):
        for j in range(i + 1, len(race.racehorses)):
            horse2 = race.racehorses[j]
            odds_dict.update(
                {
                    f"{horse1.num}-{horse2.num}": wideOdds(horse1, horse2)
                    if baken == "wide"
                    else umarenOdds(horse1, horse2)
                }
            )
    return odds_dict


def wideOdds(horse1: RacehorseData, horse2: RacehorseData):
    p = horse1.predict.pp_icchaku * horse2.predict.pp_nichaku / (
        1 - horse1.predict.pp_nichaku
    ) + horse1.predict.pp_icchaku * horse2.predict.pp_sanchaku / (
        1 - horse1.predict.pp_sanchaku
    )
    p += horse1.predict.pp_nichaku * horse2.predict.pp_icchaku / (
        1 - horse1.predict.pp_icchaku
    ) + horse1.predict.pp_nichaku * horse2.predict.pp_sanchaku / (
        1 - horse1.predict.pp_sanchaku
    )
    p += horse1.predict.pp_sanchaku * horse2.predict.pp_icchaku / (
        1 - horse1.predict.pp_icchaku
    ) + horse1.predict.pp_sanchaku * horse2.predict.pp_nichaku / (
        1 - horse1.predict.pp_nichaku
    )
    return 1 / p


def wakurenOddses(race: BangumiData):
    odds_dict = {}
    for i in range(8):
        for j in range(i, 8):
            gate1 = i + 1
            gate2 = j + 1
            odds_dict.update({f"{gate1}-{gate2}": wakurenOdds(race, gate1, gate2)})
    return odds_dict


def wakurenOdds(race: BangumiData, gate1: int, gate2: int):
    horses1 = [horse for horse in race.racehorses if horse.waku == gate1]
    if gate1 == gate2:
        return wakurenSameOdds(horses1)
    horses2 = [horse for horse in race.racehorses if horse.waku == gate2]
    pp = 0
    if len(horses1) == 0 or len(horses2) == 0:
        return 0
    for horse1 in horses1:
        for horse2 in horses2:
            pp += 1 / umarenOdds(horse1, horse2)
    return 1 / pp


def wakurenSameOdds(horses: List[RacehorseData]):
    if len(horses) <= 1:
        return 0
    pp = 0
    for i, horse1 in enumerate(horses):
        for j in range(i + 1, len(horses)):
            horse2 = horses[j]
            pp += 1 / umarenOdds(horse1, horse2)
    return 1 / pp



"""期間してい可能なローダーのインターフェース"""

from typing import List, Protocol, runtime_checkable
from jrdb_model import KaisaiData


@runtime_checkable
class SpecifiedLoader(Protocol):
    """IOロードのインターフェース"""

    start: int
    end: int

    def load(self) -> List[KaisaiData]:
        """開催データ一覧を取得する

        Returns:
            List[KaisaiData]: 開催データ一覧
        """

    @property
    def period(self) -> str:
        """期間にアクセス可能"""



import numpy as np


class CategoryGetter:
    # 開催データ
    def getKaisaikbn(self, value):
        return self.categories(3, value)

    def getDayofweek(self, value):
        types = {"日": 1, "月": 2, "火": 3, "水": 4, "木": 5, "金": 6, "土": 7}
        return self.categories(7, types[value])

    def getTenko(self, value):
        return self.categories(6, value)

    def getBabaabst(self, value):
        return self.categories(4, value)

    def getBabadetail(self, baba):
        return self.categories(3, baba)

    def getTurfkind(self, kind):
        return self.categories(3, kind)

    def getTennatsu(self, ten):
        return self.categories(2, ten)

    def getStopfreeze(self, st):
        return self.categories(2, st)

    # 番組データ
    def getDistance(self, value):
        if 2400 < value <= 2500:
            return self.categories(12, 9)
        if 2500 < value <= 3000:
            return self.categories(12, 10)
        if 3000 < value <= 3200:
            return self.categories(12, 11)
        if value > 3200:
            return self.categories(12, 12)
        return self.categories(12, (value // 200) - 4)

    def getTdscode(self, tds):
        return self.categories(3, tds)

    def getRightleft(self, rl):
        return self.categories(4, rl)

    def getInout(self, inout):
        return self.categories(4, inout)

    def getShubetsu(self, s):
        return self.categories(5, s)

    def getJoken(self, value):
        types = ["04", "05", "08", "09", "10", "15", "16", "A1", "A2", "A3", "OP"]
        return self.categories(11, types.index(value) + 1)

    def getJuryo(self, j):
        return self.categories(4, j)

    def getGrade(self, g):
        return self.categories(5, g)

    def getCourse(self, c):
        return self.categories(5, c)

    # RaceHorse
    # 汎用メソッド利用
    def getNum(self, num):
        return self.categories(18, num)

    def getLegtype(self, leg_type):
        return self.categories(7, int(leg_type))

    def getOmotekiseicode(self, omotekisei_code):
        return self.categories(4, omotekisei_code)

    def getBrinkers(self, brinkers):
        return self.categories(4, brinkers)

    def getMinarai(self, minarai):
        return self.categories(4, minarai)

    def getWaku(self, waku):
        return self.categories(8, waku)

    def getTurfadjustcode(self, turf_adjust_code):
        return self.categories(5, turf_adjust_code)

    def getTorikeshi(self, torikeshi):
        return self.categories(2, torikeshi)

    def getSex(self, sex):
        return self.categories(3, sex)

    def getBanushikaicode(self, banushikai_code):
        return self.categories(11, banushikai_code)

    def getYuso(self, yuso):
        return self.categories(5, yuso)

    def getKokyuflg(self, kokyu_flg):
        return self.categories(3, kokyu_flg)

    def getTurfdartsteepleflg(self, turf_dart_steeple_flg):
        return self.categories(3, turf_dart_steeple_flg)

    def getDistanceflg(self, distance_flg):
        return self.categories(2, distance_flg)

    def getClassflg(self, class_flg):
        return self.categories(4, class_flg)

    def getTenkyuflg(self, tenkyu_flg):
        return self.categories(4, tenkyu_flg)

    def getKyoseiflg(self, kyosei_flg):
        return self.categories(4, kyosei_flg)

    def getTraincoursekind(self, train_course_kind):
        return self.categories(6, train_course_kind)

    def getSaka(self, saka):
        return self.categories(2, saka)

    def getWood(self, wood):
        return self.categories(2, wood)

    def getDart(self, dart):
        return self.categories(2, dart)

    def getTurf(self, turf):
        return self.categories(2, turf)

    def getPool(self, pool):
        return self.categories(2, pool)

    def getSteeple(self, steeple):
        return self.categories(2, steeple)

    def getPolitruck(self, politruck):
        return self.categories(2, politruck)

    def getTraindistance(self, train_distance):
        return self.categories(5, train_distance)

    def getTrainjuten(self, train_juten):
        return self.categories(5, train_juten)

    def getOikirikind(self, oikiri_kind):
        return self.categories(4, oikiri_kind)

    def getZensoIjokbn(self, zenso_ijo_kbn):
        return self.categories(7, zenso_ijo_kbn)

    def getZensoCourseposition(self, zenso_course_position):
        return self.categories(9, zenso_course_position)

    def getZensoRacelegtype(self, zenso_race_leg_type):
        return self.categories(7, zenso_race_leg_type)

    def getZensoCorner4courseposition(self, zenso_corner4_course_position):
        return self.categories(8, zenso_corner4_course_position)

    def getUmakigocode(self, uk):
        return self.categories(28, uk)

    # 独自メソッド利用
    def getBacode(self, value):
        if (value < 1) or (value > 10):
            value = 11
        return self.categories(11, value)

    def getDistanceadjust(self, value):
        if 0 <= value <= 6:
            value = value + 1
        else:
            value = 8
        return self.categories(8, value)

    def getJockeyshirushi(self, value):
        if 0 <= value <= 6:
            value = value + 1
        else:
            value = 8
        return self.categories(8, value)

    def getDartadjustcode(self, d):
        if 0 <= value <= 3:
            value = value + 1
        else:
            value = 5
        return self.categories(5, value)

    def getGekisotype(self, value):
        types = {"00": 1, "A1": 2, "A2": 3, "A3": 4, "A4": 5, "B1": 6, "B2": 7}
        return self.categories(7, types[value])

    def getRestreasoncode(self, value):
        if 0 <= value <= 7:
            return self.categories(11, value + 1)
        if 11 <= value <= 16:
            return self.categories(11, value - 2)
        if value == 21:
            return self.categories(11, 15)

    def getNorikaeflg(self, value):
        if 0 <= value <= 1:
            value = value + 1
        else:
            value = 3
        return self.categories(3, value)

    def getHobokusakirank(self, value):
        types = {"0": 1, "A": 2, "B": 3, "C": 4, "D": 5, "E": 6}
        return self.categories(6, types[value])

    def getTraintype(self, value):
        return self.categories(11, int(value))

    def getTrainvolhyoka(self, value):
        types = {"A": 1, "B": 2, "C": 3, "D": 4, " ": 5}
        return self.categories(5, types[value])

    # 汎用メソッド

    def categories(self, category_count, value):
        cat = np.zeros([category_count])
        cat[value - 1] = 1
        return cat



"""トレーニングに利用する関数."""

from functools import reduce
from typing import List

import numpy as np
from jrdb_model import BangumiData, KaisaiData, RacehorseData
from sklearn.preprocessing import StandardScaler


def create_score_data_matrix(kaisais: List[KaisaiData]) -> np.ndarray:
    """開催データからトレーニング用データマトリクスを作成する.

    Args:
        kaisais (List[KaisaiData]): 開催データ

    Returns:
        np.ndarray: 数値データ行列[レース, 馬, 特徴量]

    """
    base_matrix = np.zeros(
        [
            reduce(lambda x, y: x + len(y.races), kaisais, 0),
            18,
            number_of_score_features(kaisais[0]),
        ]
    )
    return _set_scores(base_matrix, kaisais)


def _set_scores(score_data: np.ndarray, kaisais: List[KaisaiData]) -> np.ndarray:
    w = 0
    for kaisai in kaisais:
        for race in kaisai.races:
            score_data = _set_race_scores(score_data, race, w, _kaisai_scores(kaisai))
            w = w + 1
    return score_data


def _set_race_scores(
    matrix: np.ndarray, race: BangumiData, race_num: int, kaisai_scores
) -> np.ndarray:
    scores = kaisai_scores
    scores += _race_scores(race)
    for horse in race.racehorses:
        matrix = _set_score_data(
            matrix,
            race_num,
            horse.num - 1,
            _filter_string_to_int(scores + _horse_scores(horse)),
        )
    return matrix


def standardize(matrix: np.ndarray) -> np.ndarray:
    """行列に標準化処理を施す.

    Args:
        matrix (np.ndarray): 数値データ行列_

    Returns:
        np.ndarray: 標準化済み行列

    """
    sds = matrix
    for i, element in enumerate(matrix):
        ss = StandardScaler()
        ss.fit(element)
        sds[i] = ss.transform(element)
    return sds


def number_of_score_features(kaisai: KaisaiData) -> int:
    """開催データから特徴量の数量を取得する.

    Args:
        kaisai (KaisaiData): 開催データ

    Returns:
        int: 特徴量の数

    """
    return len(
        _kaisai_scores(kaisai)
        + _race_scores(kaisai.races[0])
        + _horse_scores(kaisai.races[0].racehorses[0])
    )


def _set_score_data(
    matrix: np.ndarray, race_num: int, horse_num: int, scores: List
) -> np.ndarray:
    features = len(matrix[race_num, horse_num])
    if features != len(scores):
        raise RuntimeError("特徴量の数と行列の次元数が一致しません")
    for s in range(features):
        matrix[race_num, horse_num, s] = scores[s]
    return matrix


def _kaisai_scores(kaisai: KaisaiData) -> List:
    return [
        kaisai.turf_baba_in,
    ]


def _race_scores(race: BangumiData):
    return []


def _horse_scores(horse: RacehorseData) -> List:
    return [
        horse.idm,
        horse.jockey_score,
        horse.info_score,
        horse.train_score,
        horse.trainer_score,
        horse.train_code,
        horse.jockey_rate_rentai,
        horse.kakutoku_money,
        horse.ls_score_order,
        horse.up_score_order,
        horse.expect_jokey_win_rate,
        horse.expect_jokey_rentai_rate,
        horse.calculated_score.waku_win_rate,
        horse.calculated_score.waku_rentai_rate,
    ]


def _filter_string_to_int(scores: List) -> List[int]:
    return [0 if isinstance(score, str) else score for score in scores]



"""ラベルを作成する"""

from typing import List
from functools import reduce
import pickle
import numpy as np
from jrdb_model import KaisaiData


class LabelCreator:
    """ラベル作成クラス"""

    def __init__(self, kaisais: List[KaisaiData]):
        self.kaisais = kaisais
        num_race = reduce(lambda x, y: x + len(y.races), kaisais, 0)
        num_max_horse = 18
        self.t_icchaku = np.zeros([num_race, num_max_horse])
        self.t_nichaku = np.zeros([num_race, num_max_horse])
        self.t_sanchaku = np.zeros([num_race, num_max_horse])
        self.t_yonchaku = np.zeros([num_race, num_max_horse])
        self.t_gochaku = np.zeros([num_race, num_max_horse])
        w_num = 0
        for kaisai in kaisais:
            for race in kaisai.races:
                for horse in race.racehorses:
                    oa = horse.result.order_of_arrival
                    if oa == 1:
                        self.t_icchaku[w_num, horse.num - 1] = 1
                    elif oa == 2:
                        self.t_nichaku[w_num, horse.num - 1] = 1
                    elif oa == 3:
                        self.t_sanchaku[w_num, horse.num - 1] = 1
                    elif oa == 4:
                        self.t_yonchaku[w_num, horse.num - 1] = 1
                    elif oa == 5:
                        self.t_gochaku[w_num, horse.num - 1] = 1
                w_num += 1
        self.labels = [
            self.t_icchaku,
            self.t_nichaku,
            self.t_sanchaku,
            self.t_yonchaku,
            self.t_gochaku,
        ]

    def save(self, name: str):
        """ラベルデータをローカルに保存する

        Args:
            name (str): ファイル名
        """
        with open(f"{name}.dump", mode="wb") as f:
            pickle.dump(self.labels, f)
